#! /usr/bin/python
# -*- coding: utf-8 -*-

# HARK Middleware
# Copyright (C) 2013-2019 by Honda Research Institute Japan Co., Ltd.
# $Id:$

import logging
import logging.config
import os

logger = logging.getLogger('hark-middleware')


def initialize_log(
        config_file=None,
        log_level='info',
        output_file=None,
        output_fd=None):
    global logger

    if config_file:
        logging.config.fileConfig(config_file)
    if output_file is not None:
        setup_default(log_level, logging.FileHandler(output_file))
    if output_fd is not None:
        setup_default(log_level, FDHandler(output_fd))
    if not config_file and output_file is None and output_fd is None:
        setup_default(log_level, logging.StreamHandler())

    logger = logging.getLogger('hark-middleware')

    # # 'application' code
    # logger.debug('debug message')
    # logger.info('info message')
    # logger.warning('warn message')
    # logger.error('error message')
    # logger.critical('critical message')


def setup_default(log_level, ch):
    global logger

    log_level = log_level.upper()
    if log_level == 'DEBUG':
        level = logging.DEBUG
    elif log_level == 'INFO':
        level = logging.INFO
    elif log_level in ('WARN', 'WARNING'):
        level = logging.WARNING
    elif log_level == 'ERROR':
        level = logging.ERROR
    elif log_level == 'CRITICAL':
        level = logging.CRITICAL
    else:
        raise BaseException('Unexpected log level value: {}'.format(log_level))

    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(process)d - %(levelname)s - %(message)s')
    ch.setLevel(level)
    ch.setFormatter(formatter)

    logger = logging.getLogger('hark-middleware')
    logger.setLevel(level)
    logger.addHandler(ch)


class FDHandler(logging.StreamHandler):
    def __init__(self, fd, mode='w', encoding=None, delay=False, errors=None):
        """
        Open the specified file and use it as the stream for logging.
        """
        self.fd = fd
        self.mode = mode
        self.encoding = encoding
        self.errors = errors
        self.delay = delay
        if delay:
            # We don't open the stream, but we still need to call the
            # Handler constructor to set level, formatter, lock etc.
            logging.Handler.__init__(self)
            self.stream = None
        else:
            logging.StreamHandler.__init__(self, self._open())

    def _open(self):
        return os.fdopen(
            self.fd,
            self.mode,
            encoding=self.encoding,
            errors=self.errors)

    def __repr__(self):
        level = str(self.level)
        return '<%s %d (%s)>' % (self.__class__.__name__, self.fd, level)


# end of file
