/*
 */

#include "libharkmw.h"

#include <hark-base/BufferedNode.h>
#include <hark-base/Vector.h>
#include <hark-base/Matrix.h>
#include <hark-base/Map.h>
#include <hark-base/Source.h>

#include "type.h"

///* From harkfd/Source.h */
//namespace HARK{
//    class Source : public HARK::Object {
//    public:
//        float x[3];
//        float power;
//        int id;
//        float remaining;
//
//        Source();
//        Source(const Source& src);
//        float angle(const float* y) const;
//        float angle(const Source& src);
//        virtual HARK::ObjectRef clone();
//        virtual void printOn(std::ostream& out) const;
//        void readFrom(std::istream &in);
//        friend std::istream& operator>>(std::istream &in, Source& src);
//        ~Source(){}
//    };
//}
///* From harkfd/Source.h [end] */

namespace HARKMW
{
    PyObject *
    toPythonObject(HARK::ObjectRef &obj)
    {
      // {
      //   std::cout << "Debug: toPythonObject: " << obj << std::endl;
      // }
       // fprintf(stdout, "Type: %s\n", typeid(*obj).name());
       // fprintf(stdout, "  ? %s\n", typeid(HARK::Vector<HARK::ObjectRef>).name());
       // fprintf(stdout, "  ? %s\n", typeid(HARK::Map<int, HARK::ObjectRef>).name());
       // fprintf(stdout, "  ? %s\n", typeid(HARK::Matrix<float>).name());
       // fprintf(stdout, "  ? %s\n", typeid(HARK::Matrix<HARK::Complex<float> >).name());
       // // fprintf(stdout, "  ? %s\n", typeid(HARK::Source).name());
       // fprintf(stdout, "  ? %s\n", typeid(bool).name());
       // fprintf(stdout, "  ? %s\n", typeid(HARK::NetCType<bool>).name());
       // fprintf(stdout, "  ? %s\n", typeid(HARK::NetCType<HARK::Bool>).name());
       // fprintf(stdout, "  ? %s\n", typeid(HARK::String).name());
       // fprintf(stdout, "  ? %s\n", typeid(std::string).name());
       if(typeid(*obj) == typeid(bool) ||
          typeid(*obj) == typeid(HARK::NetCType<bool>) ||
          typeid(*obj) == typeid(HARK::NetCType<HARK::Bool>)){

           if(HARK::dereference_cast<bool>(obj)){
//                fprintf(stdout, "Returning true\n");
               Py_RETURN_TRUE;
           }else{
//                fprintf(stdout, "Returning false\n");
               Py_RETURN_FALSE;
           }

         }else if(typeid(*obj) == typeid(int) ||
              typeid(*obj) == typeid(HARK::NetCType<int>) ||
              typeid(*obj) == typeid(HARK::NetCType<HARK::Int>)){

                int v = HARK::dereference_cast<int>(obj);
                return PyInt_FromLong(v);

          }else if(typeid(*obj) == typeid(float) ||
               typeid(*obj) == typeid(HARK::NetCType<float>) ||
               typeid(*obj) == typeid(HARK::NetCType<HARK::Float>)){

                 float v = HARK::dereference_cast<float>(obj);
                 return PyFloat_FromDouble(v);


           }else if(typeid(*obj) == typeid(double) ||
                typeid(*obj) == typeid(HARK::NetCType<double>) ||
                typeid(*obj) == typeid(HARK::NetCType<HARK::Double>)){

                  double v = HARK::dereference_cast<double>(obj);
                  return PyFloat_FromDouble(v);

        }else if(typeid(*obj) == typeid(std::string)){
            return PyStr_FromString(HARK::dereference_cast<std::string>(obj).c_str());

        }else if(typeid(*obj) == typeid(HARK::String)){
            return PyStr_FromString(static_cast<HARK::String *>(&(*obj))->c_str());

//        }else if(typeid(*obj) == typeid(HARK::Source)){
//            HARK::Source *value = static_cast<HARK::Source *>(&(*obj));
//            return Py_BuildValue("{sssfsfsfsfsisf}",
//                                 "_type", "HARK::Source",
//                                 "x", value->x[0],
//                                 "y", value->x[1],
//                                 "z", value->x[2],
//                                 "power", value->power,
//                                 "id", value->id,
//                                 "remaining", value->remaining);

        }else if(typeid(*obj) == typeid(HARK::Vector<HARK::ObjectRef>)){
            HARK::Vector<HARK::ObjectRef> &value = *(static_cast<HARK::Vector<HARK::ObjectRef> *>(&(*obj)));
            PyObject *pobj = PyList_New(value.size());

            size_t i = 0;
            for(HARK::Vector<HARK::ObjectRef>::iterator it = value.begin(); it != value.end(); ++it){
                HARK::ObjectRef &v = *it;
                PyList_SET_ITEM(pobj, i++, toPythonObject(v));
            }

            return pobj;

        }else if(typeid(*obj) == typeid(HARK::Matrix<float>)){
            HARK::Matrix<float> &value = *(static_cast<HARK::Matrix<float> *>(&(*obj)));
            PyObject *pobj = PyList_New(value.nrows());

            size_t i;
            for(i = 0; i < value.nrows(); i++){
                PyObject *robj = PyList_New(value.ncols());
                size_t j;
                for(j = 0; j < value.ncols(); j++){
                    PyList_SET_ITEM(robj, j, Py_BuildValue("f", value(i, j)));
                }
                PyList_SET_ITEM(pobj, i, robj);
            }

            return pobj;

        }else if(typeid(*obj) == typeid(HARK::Matrix<std::complex<float> >)){
            HARK::Matrix<std::complex<float> > &value = *(static_cast<HARK::Matrix<std::complex<float> > *>(&(*obj)));
            PyObject *pobj = PyList_New(value.nrows());

            size_t i;
            for(i = 0; i < value.nrows(); i++){
                PyObject *robj = PyList_New(value.ncols());
                size_t j;
                for(j = 0; j < value.ncols(); j++){
                    PyList_SET_ITEM(robj, j, Py_BuildValue("(ff)", value(i, j).real(), value(i, j).imag()));
                }
                PyList_SET_ITEM(pobj, i, robj);
            }

            return pobj;

        }else if(typeid(*obj) == typeid(HARK::Matrix<HARK::Complex<float> >)){
            HARK::Matrix<HARK::Complex<float> > &value = *(static_cast<HARK::Matrix<HARK::Complex<float> > *>(&(*obj)));
            PyObject *pobj = PyList_New(value.nrows());

            size_t i;
            for(i = 0; i < value.nrows(); i++){
                PyObject *robj = PyList_New(value.ncols());
                size_t j;
                for(j = 0; j < value.ncols(); j++){
                    PyList_SET_ITEM(robj, j, Py_BuildValue("(ff)", value(i, j).real(), value(i, j).imag()));
                }
                PyList_SET_ITEM(pobj, i, robj);
            }

            return pobj;

//        }else if(typeid(*obj) == typeid(HARK::Map<int, HARK::ObjectRef>)){
//            Py_RETURN_NONE;
        }else if(typeid(*obj) == typeid(HARK::Source)){
          HARK::Source &value = *(static_cast<HARK::Source *>(&(*obj)));

          PyObject *strSource = PyUnicode_FromString("Source");
          PyObject *lstX      = PyList_New(0);
          PyObject *dblPower  = PyFloat_FromDouble(value.power);
          PyObject *intId     = PyLong_FromLong(value.id);

          PyList_Append(lstX, PyFloat_FromDouble(value.x[0]));
          PyList_Append(lstX, PyFloat_FromDouble(value.x[1]));
          PyList_Append(lstX, PyFloat_FromDouble(value.x[2]));

          PyObject *p = PyDict_New();
          PyDict_SetItemString(p, "x",   lstX);
          PyDict_SetItemString(p, "pow", dblPower);
          PyDict_SetItemString(p, "id",  intId);

          PyObject *o = PyDict_New();
          PyDict_SetItemString(o, "type", strSource);
          PyDict_SetItemString(o, "value", p);

          return o;

        }else{
            std::stringstream ss;
            ss << "Failed to convert C++ object to Python object: " << std::endl;
            obj->printOn(ss);

            PyErr_SetString(PyExc_TypeError, ss.str().c_str());
            return NULL;
        }
    }

    HARK::ObjectRef
    fromPythonObject(PyObject *obj)
    {
      // {
      //   PyObject *str = PyObject_Repr(obj);
      //   std::cout << "Debug: fromPythonObject: " << PyStr_AsUTF8(str) << std::endl;
      //   Py_DECREF(str);
      // }

        HARK::ObjectRef o = HARK::nilObject;

        if(PyStr_Check(obj)){
            o = HARK::ObjectRef(new HARK::String(PyStr_AsUTF8(obj)));

        }else if(PyLong_Check(obj)){
            o = HARK::ObjectRef(new HARK::NetCType<int>(PyLong_AsLong(obj)));

        }else if(PyFloat_Check(obj)){
            o = HARK::ObjectRef(new HARK::NetCType<float>(PyFloat_AsDouble(obj)));

        }else if(PyBool_Check(obj)){
            o = HARK::ObjectRef(new HARK::Bool(obj == Py_True));

        }else if(PyList_Check(obj)){
            size_t list_size = PyList_Size(obj);

            if(list_size == 0){
              o = HARK::ObjectRef(new HARK::Vector<HARK::ObjectRef>());

            }else{
              PyObject *list_head = PyList_GetItem(obj, 0);

              if(PyDict_Check(list_head)){
                o = HARK::ObjectRef(new HARK::Vector<HARK::ObjectRef>());
                HARK::Vector<HARK::ObjectRef> &v = HARK::object_cast<HARK::Vector<HARK::ObjectRef> >(o);

                for(size_t i = 0; i < list_size; i++){
                  PyObject *list_item = PyList_GetItem(obj, i);

                  v.push_back(fromPythonObject(list_item));

                  // Py_DECREF(list_item);
                }

              }else if(PyList_Check(list_head)){
                  size_t list2_size = PyList_Size(list_head);
                  PyObject *list2_head = PyList_GetItem(list_head, 0);

                  if(PyTuple_Check(list2_head)){
                      // May be Matrix<Complex<float>>
                      HARK::Matrix<std::complex<float> > *mat = new HARK::Matrix<std::complex<float> >(list_size, list2_size);

                      int i, j, n, m;
                      for(i = 0, n = PyList_Size(obj); i < n; i++){

                          PyObject *p = PyList_GetItem(obj, i);
                          for(j = 0, m = PyList_Size(p); j < m; j++){
                              float vr, vi;
                              PyArg_Parse(PyList_GetItem(p, j), "(ff)", &vr, &vi);
                              (*mat)(i, j) = std::complex<float>(vr, vi);
                          }
                      }

                      o = HARK::ObjectRef(mat);

                  }else{
                      // May be Matrix<float>
                      HARK::Matrix<float> *mat = new HARK::Matrix<float>(list_size, list2_size);

                      int i, j, n, m;
                      for(i = 0, n = PyList_Size(obj); i < n; i++){

                          PyObject *p = PyList_GetItem(obj, i);
                          for(j = 0, m = PyList_Size(p); j < m; j++){
                              float v;
                              PyArg_Parse(PyList_GetItem(p, j), "f", &v);
                              (*mat)(i, j) = v;
                          }
                      }

                      o = HARK::ObjectRef(mat);
                  }

            }else if(PyTuple_Check(list_head)){
                // May be Complex<float>
                o = HARK::ObjectRef(new HARK::Complex<float>(list_size));

            }else{
                o = HARK::ObjectRef(new HARK::String(PyStr_AsUTF8(obj)));
            }
          }

        }else if(PyDict_Check(obj)){
          if(PyDict_Size(obj) == 0){
            return o;
          }
          PyObject *strType = PyDict_GetItemString(obj, "type");
          if(strType == NULL){
            PyErr_SetString(PyExc_TypeError, "Unexpected dict object: does not have 'type' field");

          }else{
            const std::string type = PyStr_AsUTF8(strType);
            // std::cout << "fromPythonObject: type=" << type << std::endl;

            if(type == "Source"){
              PyObject *value = PyDict_GetItemString(obj, "value");
              if(value == NULL){
                PyErr_SetString(PyExc_TypeError, "Unexpected dict object: type=\"Source\": does not have 'value' field");

              }else{
                PyObject *lstX     = PyDict_GetItemString(value, "x");
                PyObject *dblPower = PyDict_GetItemString(value, "pow");
                PyObject *intID    = PyDict_GetItemString(value, "id");
                if(lstX == NULL || dblPower == NULL || intID == NULL){
                  PyErr_SetString(PyExc_TypeError, "Unexpected dict object: type=\"Source\": requres \"x\", \"pow\", \"id\" fields in \"value\" field");

                }else{
                  o = HARK::ObjectRef(new HARK::Source());
                  HARK::Source &source = HARK::object_cast<HARK::Source>(o);

                  if(!PyList_Check(lstX)){
                    PyErr_SetString(PyExc_TypeError, "Unexpected dict object: type=\"Source\": type mismatch in \"x\"");

                  }else if(!PyFloat_Check(dblPower)){
                    PyErr_SetString(PyExc_TypeError, "Unexpected dict object: type=\"Source\": type mismatch in \"pow\"");

                  }else if(!PyInt_Check(intID)){
                    PyErr_SetString(PyExc_TypeError, "Unexpected dict object: type=\"Source\": type mismatch in \"id\"");

                  }else if(PyList_Size(lstX) != 3){
                    PyErr_SetString(PyExc_TypeError, "Unexpected dict object: type=\"Source\": type mismatch in \"x\"");

                  }else{
                    PyObject *x0 = PyList_GetItem(lstX, 0);
                    PyObject *x1 = PyList_GetItem(lstX, 1);
                    PyObject *x2 = PyList_GetItem(lstX, 2);

                    if(!PyFloat_Check(x0)){
                      PyErr_SetString(PyExc_TypeError, "Unexpected dict object: type=\"Source\": type mismatch in \"x[0]\"");
                    }else if(!PyFloat_Check(x1)){
                      PyErr_SetString(PyExc_TypeError, "Unexpected dict object: type=\"Source\": type mismatch in \"x[1]\"");
                    }else if(!PyFloat_Check(x2)){
                      PyErr_SetString(PyExc_TypeError, "Unexpected dict object: type=\"Source\": type mismatch in \"x[2]\"");
                    }else{
                      source.x[0] = PyFloat_AsDouble(x0);
                      source.x[1] = PyFloat_AsDouble(x1);
                      source.x[2] = PyFloat_AsDouble(x2);
                      source.power = PyFloat_AsDouble(dblPower);
                      source.id    = PyInt_AsLong(intID);
                    }

                    // Py_DECREF(x0);
                    // Py_DECREF(x1);
                    // Py_DECREF(x2);
                  }

                  // if(lstX){
                  //   Py_DECREF(lstX);
                  // }
                  // if(dblPower){
                  //   Py_DECREF(dblPower);
                  // }
                  // if(intID){
                  //   Py_DECREF(intID);
                  // }
                }

                // Py_DECREF(value);
              }

            }else{
              std::stringstream ss;
              ss << "Unexpected dict object with type: " << type;
              PyErr_SetString(PyExc_TypeError, ss.str().c_str());
            }

            // Py_DECREF(strType);
          }

        }else{
            std::stringstream ss;

            PyObject *str = PyObject_Repr(obj);
            ss << "Failed to convert Python object to C++ object: " << std::endl;
            ss << PyStr_AsUTF8(str);
            Py_DECREF(str);

            PyErr_SetString(PyExc_TypeError, ss.str().c_str());
        }

        return o;
    }

}

/* end of file */
